==> unity-ads-4.12.1-sources.jar: com/unity3d/ads/core/data/manager/TransactionEventManager.kt <==
package com.unity3d.ads.core.data.manager

import com.google.protobuf.Timestamp
import com.unity3d.ads.core.data.datasource.ByteStringDataSource
import com.unity3d.ads.core.data.repository.TransactionEventRepository
import com.unity3d.ads.core.domain.events.GetTransactionData
import com.unity3d.ads.core.domain.events.GetTransactionRequest
import com.unity3d.services.store.StoreMonitor
import com.unity3d.services.store.gpbl.BillingResultResponseCode
import com.unity3d.services.store.gpbl.bridges.BillingResultBridge
import com.unity3d.services.store.gpbl.bridges.PurchaseBridge
import com.unity3d.services.store.gpbl.listeners.BillingInitializationListener
import gatewayprotocol.v1.TransactionEventRequestOuterClass.TransactionData
import kotlinx.coroutines.CompletableDeferred
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.awaitAll
import kotlinx.coroutines.launch
import kotlin.coroutines.resume
import kotlin.coroutines.resumeWithException
import kotlin.coroutines.suspendCoroutine

// Deals with querying purchases from the Google Play Billing Library.
class TransactionEventManager(
    private val scope: CoroutineScope,
    private val storeMonitor: StoreMonitor,
    private val getTransactionData: GetTransactionData,
    private val getTransactionRequest: GetTransactionRequest,
    private val transactionEventRepository: TransactionEventRepository,
    private val iapTransactionStore: ByteStringDataSource
) {
    operator fun invoke() {
        scope.launch {
            if (!storeMonitor.isInitialized) {
                suspendCoroutine {
                    storeMonitor.initialize(object : BillingInitializationListener {
                        override fun onIsAlreadyInitialized() {
                            it.resume(Unit)
                        }

                        override fun onBillingSetupFinished(billingResult: BillingResultBridge) {
                            if (billingResult.responseCode != BillingResultResponseCode.OK) {
                                it.resumeWithException(Exception("Billing setup failed"))
                                return
                            }
                            it.resume(Unit)
                        }

                        override fun onBillingServiceDisconnected() {
                            it.resumeWithException(Exception("Billing service disconnected"))
                        }

                        override fun onPurchaseUpdated(
                            billingResult: BillingResultBridge,
                            purchases: List<PurchaseBridge>?
                        ) {
                            onPurchasesReceived(billingResult, purchases)
                        }
                    })
                }
            }

            storeMonitor.getPurchases(dummyOperationId, INAPP) { billingResult, purchases ->
                onPurchasesReceived(billingResult, purchases)
            }
        }

    }

    private fun onPurchasesReceived(billingResult: BillingResultBridge, purchases: List<PurchaseBridge>?) {
        if (billingResult.responseCode != BillingResultResponseCode.OK || purchases.isNullOrEmpty()) return
        scope.launch {
            val transactionDataList = mutableListOf<TransactionData>()
            val deferredPurchaseList = purchases.map { CompletableDeferred<Unit>() }
            purchases.forEachIndexed { index, purchase ->
                val purchaseTime = purchase.originalJson["purchaseTime"] as Long // Since Epoch, in ms
                val lastIapTransactionSent = Timestamp.parseFrom(iapTransactionStore.get().data)

                if (lastIapTransactionSent.seconds < purchaseTime / 1000f) {
                    storeMonitor.getSkuDetails(dummyOperationId, INAPP, listOf(purchase.originalJson["productId"] as String)) { billingResult, skuDetailsList ->
                        if (skuDetailsList == null || billingResult.responseCode != BillingResultResponseCode.OK) {
                            deferredPurchaseList[index].complete(Unit)
                            return@getSkuDetails
                        }
                        transactionDataList.addAll(skuDetailsList.map { getTransactionData(purchase, it) })
                        deferredPurchaseList[index].complete(Unit)
                    }
                } else {
                    deferredPurchaseList[index].complete(Unit)
                }
            }

            // Wait for all purchases to complete
            awaitAll(*deferredPurchaseList.toTypedArray())

            if (transactionDataList.isNotEmpty()) {
                val transactionRequest = getTransactionRequest(transactionDataList)
                transactionEventRepository.addTransactionEvent(transactionRequest)
            }
        }
    }

    companion object {
        private const val dummyOperationId = 42; // Operation ID not used outside of WebView context.
        private const val INAPP = "inapp"
    }
}

==> unity-ads-4.12.2-sources.jar: com/unity3d/ads/core/data/manager/TransactionEventManager.kt <==
package com.unity3d.ads.core.data.manager

import com.google.protobuf.Timestamp
import com.unity3d.ads.core.data.datasource.ByteStringDataSource
import com.unity3d.ads.core.data.model.exception.TransactionException
import com.unity3d.ads.core.data.repository.TransactionEventRepository
import com.unity3d.ads.core.domain.SendDiagnosticEvent
import com.unity3d.ads.core.domain.SendDiagnosticEvent.Companion.BILLING_SERVICE_UNAVAILABLE
import com.unity3d.ads.core.domain.events.GetTransactionData
import com.unity3d.ads.core.domain.events.GetTransactionRequest
import com.unity3d.services.store.StoreMonitor
import com.unity3d.services.store.gpbl.BillingResultResponseCode
import com.unity3d.services.store.gpbl.bridges.BillingResultBridge
import com.unity3d.services.store.gpbl.bridges.PurchaseBridge
import com.unity3d.services.store.gpbl.listeners.BillingInitializationListener
import gatewayprotocol.v1.TransactionEventRequestOuterClass.TransactionData
import kotlinx.coroutines.CompletableDeferred
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.awaitAll
import kotlinx.coroutines.launch
import kotlinx.coroutines.suspendCancellableCoroutine
import kotlin.coroutines.resume
import kotlin.coroutines.resumeWithException

// Deals with querying purchases from the Google Play Billing Library.
class TransactionEventManager(
    private val scope: CoroutineScope,
    private val storeMonitor: StoreMonitor,
    private val getTransactionData: GetTransactionData,
    private val getTransactionRequest: GetTransactionRequest,
    private val transactionEventRepository: TransactionEventRepository,
    private val iapTransactionStore: ByteStringDataSource,
    private val sendDiagnosticEvent: SendDiagnosticEvent,
) {
    operator fun invoke() {
        scope.launch {
            if (!storeMonitor.isInitialized) {
                try {
                    suspendCancellableCoroutine {
                        storeMonitor.initialize(object : BillingInitializationListener {
                            private fun tryResume() {
                                if (it.isActive) it.resume(Unit)
                            }

                            private fun tryResumeWithException(exception: Exception) {
                                if (it.isActive) it.resumeWithException(exception)
                            }

                            override fun onIsAlreadyInitialized() {
                                tryResume()
                            }

                            override fun onBillingSetupFinished(billingResult: BillingResultBridge) {
                                if (billingResult.responseCode != BillingResultResponseCode.OK) {
                                    tryResumeWithException(TransactionException("Billing setup failed"))
                                    return
                                }
                                tryResume()
                            }

                            override fun onBillingServiceDisconnected() {
                                tryResumeWithException(TransactionException("Billing service disconnected"))
                            }

                            override fun onPurchaseUpdated(
                                billingResult: BillingResultBridge,
                                purchases: List<PurchaseBridge>?
                            ) {
                                onPurchasesReceived(billingResult, purchases)
                            }
                        })
                    }

                    storeMonitor.getPurchases(dummyOperationId, INAPP) { billingResult, purchases ->
                        onPurchasesReceived(billingResult, purchases)
                    }
                } catch (e: TransactionException) {
                    sendDiagnosticEvent(BILLING_SERVICE_UNAVAILABLE)
                }
            }
        }

    }

    private fun onPurchasesReceived(billingResult: BillingResultBridge, purchases: List<PurchaseBridge>?) {
        if (billingResult.responseCode != BillingResultResponseCode.OK || purchases.isNullOrEmpty()) return
        scope.launch {
            val transactionDataList = mutableListOf<TransactionData>()
            val deferredPurchaseList = purchases.map { CompletableDeferred<Unit>() }
            purchases.forEachIndexed { index, purchase ->
                val purchaseTime = purchase.originalJson["purchaseTime"] as Long // Since Epoch, in ms
                val lastIapTransactionSent = Timestamp.parseFrom(iapTransactionStore.get().data)

                if (lastIapTransactionSent.seconds < purchaseTime / 1000f) {
                    storeMonitor.getSkuDetails(dummyOperationId, INAPP, listOf(purchase.originalJson["productId"] as String)) { billingResult, skuDetailsList ->
                        if (skuDetailsList == null || billingResult.responseCode != BillingResultResponseCode.OK) {
                            deferredPurchaseList[index].complete(Unit)
                            return@getSkuDetails
                        }
                        transactionDataList.addAll(skuDetailsList.map { getTransactionData(purchase, it) })
                        deferredPurchaseList[index].complete(Unit)
                    }
                } else {
                    deferredPurchaseList[index].complete(Unit)
                }
            }

            // Wait for all purchases to complete
            awaitAll(*deferredPurchaseList.toTypedArray())

            if (transactionDataList.isNotEmpty()) {
                val transactionRequest = getTransactionRequest(transactionDataList)
                transactionEventRepository.addTransactionEvent(transactionRequest)
            }
        }
    }

    companion object {
        private const val dummyOperationId = 42; // Operation ID not used outside of WebView context.
        private const val INAPP = "inapp"
    }
}
